#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <pthread.h>
#include <math.h>
#include "Globals.h"
#include "DEM.h"
#include "Life.h"
#include "OpenGL.h"

extern int No_Exit;
extern struct Species ***life;
//extern struct Stats *stats;
extern struct Stats_T stats_t;
extern struct image_attr *ia;
extern unsigned int Threads;
extern unsigned int Thread_Step;
extern struct Thread_I *Actions_t_info;
extern struct Thread_I *Opengl_t_info;
extern double Backup_Freq_Runtime;
extern double Year_Runtime;
extern uint32_t Backup_Freq;
extern pthread_barrier_t Barrier1;
extern pthread_barrier_t Barrier2;
//extern pthread_barrier_t Barrier3;
//extern pthread_barrierattr_t Barrier1_Attr;
//extern pthread_barrierattr_t Barrier2_Attr;
//extern pthread_barrierattr_t Barrier3_Attr;
extern char *filename_backup_load;
extern char *filename_backup_save;
extern char *filename_dem;
extern char *life_speed_up_X;
extern char *life_speed_up_Y;
extern int Stats_Handled;

void Usage(void){
	printf("lr (version 0.1.1 alpha) Usage:\n");
	printf("lr [-l filename1] [-s filename2] [-f freq] [-t threads] [-v]\n");
	printf("\t[optional]\n");
	printf("\tfilename1\t= backup file to load. Default to a fresh simulation\n");
	printf("\tfilename2\t= file to save backups to. Default: BACKUP.lr\n");
	printf("\tfreq\t= unsigned integer frequency to do backup (1K year multiple). Default 1\n");
	printf("\tthreads\t= unsigned integer number of \"action\" threads to use (1 added for gui). Default 1\n");
	printf("\t-v\t= print this menu and exit\n");
}

int main(int argc, char **argv){
	//uint64_t year = 1;
	//location to place initial tribes for fresh simulation
	//const int w = INITIAL_TRIBE_X;
	//const int h = INITIAL_TRIBE_Y;
	unsigned char rgb[3];
	unsigned int n3 = 0;
	unsigned int n1 = 0;
	unsigned int n2 = 0;
	unsigned int Tcount = 0;
	int ret_val = 0;
    filename_dem = (char *)malloc(strlen("myout.test") + 1);
    strcpy(filename_dem, "myout.test");

	if(argc > 10){
		Usage();
		exit(1);
	}
	Year_Runtime = 0;
	Backup_Freq = 1000;
	Backup_Freq_Runtime = 0;
	filename_backup_load = NULL;
	filename_backup_save = NULL;
	uint32_t *options = (uint32_t *)malloc( sizeof(uint32_t) * 5 );
	(void)memset((void *)options, '\0', sizeof(uint32_t) * 5 );
	Threads = 1;

	//atexit(Usage);
	//printf("argc %d\n", argc);
	//parse user input
	if(argc > 1){ //if atleast one option
		for(n1=1; n1<argc; n1++){
			if( strcmp( (const char *)argv[n1], "-l") == 0){
				options[0] = n1+1;
				if(n1+1 > argc-1)
					n2 = 1;
			}
			if( strcmp( (const char *)argv[n1], "-s") == 0){
				options[1] = n1+1;
				if(n1+1 > argc-1)
					n2 = 1;
			}
			if( strcmp( (const char *)argv[n1], "-f") == 0){
				options[2] = n1+1;
				if(n1+1 > argc-1)
					n2 = 1;
			}
			if( strcmp( (const char *)argv[n1], "-t") == 0){
				options[3] = n1+1;
				if(n1+1 > argc-1)
					n2 = 1;
			}
			if( strcmp( (const char *)argv[n1], "-v") == 0){
				if(n1+1 > argc-1)
					n2 = 1;
			}
		}
		if(n2){
			Usage();
			free(filename_dem);
			free(options);
			exit(1);
		}
	}

	n2 = 1;
	if(options[0] != 0){
		filename_backup_load = (char *)malloc( sizeof(char) * strlen(argv[options[0]]) + 1 );
		(void)strcpy(filename_backup_load, (const char *)argv[options[0]]);
	}else{
		filename_backup_load = NULL;
	}
	if(options[1] != 0){
		filename_backup_save = (char *)malloc( sizeof(char) * strlen(argv[options[1]]) + 1 );
		(void)strcpy(filename_backup_save, (const char *)argv[options[1]]);
	}else{
		filename_backup_save = (char *)malloc( sizeof(char) * strlen("BACKUP.lr") + 1 );
		(void)strcpy(filename_backup_save, "BACKUP.lr");
	}
	if(options[2] != 0){
		Backup_Freq = atoi(argv[options[2]]);
		if(Backup_Freq < 100){
			printf("Error: Frequency input is not a valid number!\n");
			printf("\tacceptable range 100+\n");
			n2 = 2;
		}
	}
	if(options[3] != 0){
		Threads = atoi(argv[options[3]]);
		if(Threads == 0 || Threads > HEIGHT/2){  //don't break later
			printf("Error: Threads input is not a valid number!\n");
			printf("\tacceptable range 1 to %u\n", HEIGHT/2 );
			n2 = 2;
		}
	}
	if(!n2){
		Usage();
		free(options);
		free(filename_dem);
		free(filename_backup_load);
		free(filename_backup_save);
		exit(1);
	}

	//if(filename_backup_load != NULL){
    if( Load_Terrain() ){
        printf("Exiting!\n");
		free(options);
		free(filename_dem);
		free(filename_backup_load);
		free(filename_backup_save);
        exit(1);
    }

	Thread_Step = floor(HEIGHT/Threads);
	//printf("thread_step: %u backup_freq: %u\n", Threads, Thread_Step, Backup_Freq);
	printf("Threads: %u thread_step: %u backup_freq: %u filename_load: %s filename_save: %s\n", Threads, Thread_Step, Backup_Freq, filename_backup_load, filename_backup_save);
	printf("Dem dimensions(W x H): %u x %u\n", WIDTH, HEIGHT);

	pthread_t OpenGL_thread_ptr;
	pthread_t *Actions_thread_ptrs = (pthread_t *)malloc(sizeof(pthread_t) * Threads );
	unsigned int *Actions_t_ids    = (unsigned int *)malloc( sizeof(unsigned int) * Threads );
	//unsigned int OpenGL_t_ids;
	Actions_t_info = (struct Thread_I *)malloc( sizeof(struct Thread_I) * Threads );
	//Opengl_t_info  = (struct Thread_I *)malloc( sizeof(struct Thread_I) );
	stats = (struct Stats *)malloc( sizeof(struct Stats) * Threads );

	(void)memset((void *)&stats_t, '\0', sizeof(struct Stats_T) );
	for(n1=0; n1<Threads; n1++){
		(void)memset((void *)&stats[n1], '\0', sizeof(struct Stats) );
	}

	n2 = 0;
	for(n1=0; n1<Threads; n1++){
		Actions_t_info[n1].start_h = n2;
		Actions_t_info[n1].end_h   = (n1+1 != Threads) ? (n1+1)*Thread_Step : HEIGHT;
		n2 += Thread_Step;
		//printf("start_h: %u end_h: %u\n", Actions_t_info[n1].start_h, Actions_t_info[n1].end_h);
		Actions_t_ids[n1] = n1;
	}

	//initialize empty life array
	Allocate_Life_Array();
	Drawl_Image();
    life_speed_up_Y = (char *)malloc(sizeof(char)*HEIGHT);
    life_speed_up_X = (char *)malloc(sizeof(char)*WIDTH);
    (void)memset( (void *)life_speed_up_Y, '\0', sizeof(char)*HEIGHT );
    (void)memset( (void *)life_speed_up_X, '\0', sizeof(char)*WIDTH );

	printf("Allocated Life array\n");
	n3 = 0;
	if(filename_backup_load != NULL){
		if( Load_State() ){
			printf("Exiting!\n");
			free(options);
			free(filename_dem);
			free(filename_backup_load);
			free(filename_backup_save);
			free(Actions_thread_ptrs);
			free(Actions_t_ids);
			free(Actions_t_info);
			free(life_speed_up_Y);
			free(life_speed_up_X);
			//free(Opengl_t_info);
			free(stats);
			exit(1);
		}
		for(n1=0; n1<HEIGHT; n1++){
			for(n2=0; n2<WIDTH; n2++){
				if(life[n1][n2]->alive){
					life_speed_up_Y[n1] = 1;
					life_speed_up_X[n2] = 1;
				}
			}
		}
		Adjust_Sea_Level(1);
		Adjust_Climate(1);
		//printf("\tyear: %u sealevel: %d rising: %d ", stats_t.year, stats_t.cur_sealevel, stats_t.sea_rising);
		//printf("temp: %d rising %d\n", stats_t.cur_temp, stats_t.temp_rising);
	}else{
		printf("y: lower/upper %u %u\n", INITIAL_TRIBE_X-TRIBE_SQ, INITIAL_TRIBE_X+TRIBE_SQ);
		printf("x: lower/upper %u %u\n", INITIAL_TRIBE_Y-TRIBE_SQ, INITIAL_TRIBE_Y+TRIBE_SQ);
		Adjust_Sea_Level(0);
		Adjust_Climate(0);
		printf("cur_sealevel: %d\n", stats_t.cur_sealevel);
		//"plant" first primate tribe [somewhere in africa :)]
		stats_t.total_runtime = 4; //set to 4 seconds to account for initialization
		for(n1=INITIAL_TRIBE_X-TRIBE_SQ; n1<INITIAL_TRIBE_X+TRIBE_SQ; n1++){
			for(n2=INITIAL_TRIBE_Y-TRIBE_SQ; n2<INITIAL_TRIBE_Y+TRIBE_SQ; n2++){
				//need to check terrain for water
				//printf("here a\n");
				//printf("%d, %d\n", ia->dem[n2][n1], stats_t.cur_sealevel);
				if(ia->dem[n2][n1] > stats_t.cur_sealevel){
					//printf("here b\n");
					//life[n2][n1] = (struct Species *)malloc( sizeof(struct Species) );
					Add_First_Life( life[n2][n1] );
					Gen_Tribe_Color(life[n2][n1], rgb);
					ia->image[WIDTH*n2*3 + n1*3]     = rgb[0];
					ia->image[WIDTH*n2*3 + n1*3 + 1] = rgb[1];
					ia->image[WIDTH*n2*3 + n1*3 + 2] = rgb[2];
					n3++;
                    life[n2][n1]->alive = 1;
					life_speed_up_X[n1] = 1;
					life_speed_up_Y[n2] = 1;
				}
			}
		}
		stats_t.total_living = n3;
		printf("Added %u initial tribes!\n", n3);
	}
	printf("\tyear: %u sealevel: %d rising: %d ", stats_t.year, stats_t.cur_sealevel, stats_t.sea_rising);
	printf("temp: %.02f rising %d\n", stats_t.cur_temp, stats_t.temp_rising);



	//allocate fought/mated array
	//FM_Alloc();

	No_Exit = 1;

	printf("Launching OpenGL thread.\n");
	glutInit(&argc,argv);
	if(pthread_create(&OpenGL_thread_ptr, NULL, Run_GL, NULL) != 0){
		printf("Error: pthread_create failed to launch opengl thread !\n");
		exit(1);
	}

	ret_val = pthread_barrier_init(&Barrier1, NULL, Threads);
	if(ret_val != 0){
		printf("Error: pthread_barrier_init Failed: %d\n", ret_val);
		printf("\tExiting!\n");
		free(options);
		free(filename_dem);
		free(filename_backup_load);
		free(filename_backup_save);
		free(Actions_thread_ptrs);
		free(Actions_t_ids);
		free(Actions_t_info);
		free(life_speed_up_Y);
		free(life_speed_up_X);
		//free(Opengl_t_info);
		free(stats);
		exit(1);
	}
	ret_val = pthread_barrier_init(&Barrier2, NULL, Threads);
	if(ret_val != 0){
		printf("Error: pthread_barrier_init Failed: %d\n", ret_val);
		printf("\tExiting!\n");
		free(options);
		free(filename_dem);
		free(filename_backup_load);
		free(filename_backup_save);
		free(Actions_thread_ptrs);
		free(Actions_t_ids);
		free(Actions_t_info);
		free(life_speed_up_Y);
		free(life_speed_up_X);
		//free(Opengl_t_info);
		free(stats);
		exit(1);
	}
	//pthread_barrier_init(&Barrier3, &Barrier3_Attr, Threads);

    //allocate fought/mated array
    FM_Alloc();

	Stats_Handled = 0;
	printf("Launching %u Action threads.\n", Threads);
	for(Tcount = 0; Tcount < Threads; Tcount++){
		if(pthread_create(&Actions_thread_ptrs[Tcount], NULL, Tribe_Actions, (void*)&Actions_t_ids[Tcount]) != 0){
			printf("Error: pthread_create failed for action thread: %d !\n", Tcount);
			exit(1);
		}
	}

	if(pthread_join(OpenGL_thread_ptr, NULL) != 0){
		printf("Error: pthread_join failed to join OpenGL thread!\n");
		exit(1);
	}

	for(Tcount=0; Tcount < Threads; Tcount++){
		if(pthread_join(Actions_thread_ptrs[Tcount], NULL) != 0){
			printf("Error: pthread_join failed for action thread: %d !\n", Tcount);
			exit(1);
		}
	}

	printf("Cleaning up allocated objects!\n");
	free(ia->image);
	for(n1=0; n1<HEIGHT; n1++){
		free(ia->dem[n1]);
	}
	free(ia->dem);
	free(ia);
	//free(thread_i);
	for(n1=0; n1<HEIGHT; n1++){
		for(n2=0; n2<WIDTH; n2++){
			free(life[n1][n2]);
		}
		free(life[n1]);
	}
	free(life);
	for(n1=0; n1<HEIGHT; n1++){
		free(fought[n1]);
		free(mated[n1]);
	}
	free(fought);
	free(mated);
	free(stats);
	free(filename_dem);
	free(Actions_thread_ptrs);
	free(Actions_t_info);
	free(Opengl_t_info);
	free(life_speed_up_Y);
	free(life_speed_up_X);
	//if(pthread_barrierattr_destroy(&Barrier1_Attr) != 0){
	//	printf("pthread_barrierattr_destroy(&Barrier1_Attr) Failed:\n");
	//	exit(1);
	//}
	//if(pthread_barrierattr_destroy(&Barrier2_Attr) != 0){
	//	printf("pthread_barrierattr_destroy(&Barrier2_Attr) Failed:\n");
	//	exit(1);
	//}
	//if(pthread_barrierattr_destroy(&Barrier3_Attr) != 0){
	//	printf("pthread_barrierattr_destroy(&Barrier3_Attr) Failed:\n");
	//	exit(1);
	//}
	if(pthread_barrier_destroy(&Barrier1) != 0){
		printf("pthread_barrier_destroy(&Barrier1) Failed:\n");
		exit(1);
	}
	if(pthread_barrier_destroy(&Barrier2) != 0){
		printf("pthread_barrier_destroy(&Barrier2) Failed:\n");
		exit(1);
	}
	//if(pthread_barrier_destroy(&Barrier3) != 0){
	//	printf("pthread_barrier_destroy(&Barrier3) Failed:\n");
	//	exit(1);
	//}
	return 0;
}